# !/usr/bin/python3
"""
GeoHTML (LeucipPy.GeoHTML)
------------------------------------------------------
This class manipulates given dataframes into preformatted html reports using matplotlib.
The class assists in high volume output for visual analysis
"""
import base64
import gc
import io

import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from scipy.stats import gaussian_kde
import matplotlib
matplotlib.use('Agg')


class ReportMaker:
    """Class for a single atom

    :data html_string:
    :data outpath:
    :data title:
    """

    def __init__(self, title,outpath, cols=3, remove_strip=False,clr_header='lightsteelblue',clr_background='lightblue',clr_behindplot='seashell'):
        """Initialises a GeoPdb with a biopython structure

        :param biopython_structure: A list of structures that has been created from biopython
        """

        #self.df = dataframe
        #a few of the style params for the html report
        self.clr_background = clr_background
        self.clr_behindplot = clr_behindplot
        self.clr_header = clr_header

        self.outpath = outpath
        self.title=title
        self.cols = cols
        self.num_col = 0
        self.html_string = self.getHeaderString(self.title, remove_strip)
        self.table_open = False
        self.overlay_open = False
        self.fig, self.ax = plt.subplots()

    def printReport(self):
        if self.table_open:
            self.html_string += '</table>\n'
            self.table_open = False
        self.html_string += self.getFooterString()
        f = open(self.outpath, "w+")
        f.write(self.html_string)
        f.close()

    def changeColNumber(self,cols):
        self.cols = cols
        if self.table_open:
            self.html_string += '</table>\n'
            self.table_open = False
            self.num_col = 0

    def addLineComment(self,comment):
        if self.table_open:
            self.html_string += '</tr></table>\n'
            self.table_open = False
            self.num_col=0
        self.html_string += '<p>'+comment+'</p>\n'

    def addBoxComment(self,comment):
        self.HTMLIncrement()
        self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>'+comment+'</td>\n'

    def addPlot2d(self,data,plottype,geo_x,geo_y,hue,title='',palette='viridis',overlay=False,alpha=1,xrange=[None,None],yrange=[None,None],crange=[None,None]):
        self.incrementOverlay(overlay)
        self.ax.grid(which='major', color='Gainsboro', linestyle='-')
        self.ax.set_axisbelow(True)

        xarange = [xrange[0],xrange[1]] # by ref obs need to be copied
        yarange = [yrange[0],yrange[1]]
        carange = [crange[0],crange[1]]

        count = len(data[geo_x])
        minx,maxx = 0,0
        miny, maxy = 0, 0
        cmin, cmax = 0, 0
        try:
            minx, maxx = min(data[geo_x]), max(data[geo_x])
            miny, maxy = min(data[geo_y]), max(data[geo_y])
            cmin, cmax = min(data[hue]), max(data[hue])
        except:
            pass
        if xarange[0] != None:
            minx=xarange[0]
        if xarange[1] != None:
            maxx=xarange[1]
        if yarange[0] != None:
            miny=yarange[0]
        if yarange[1] != None:
            maxy=yarange[1]
        if carange[0] != None:
            cmin=carange[0]
        if carange[1] != None:
            cmax=carange[1]

        if plottype == 'scatter':
            g = self.ax.scatter(data[geo_x], data[geo_y], c=data[hue], cmap=palette, edgecolor='silver', alpha=alpha, linewidth=0.5, s=20,vmin=cmin,vmax=cmax)
            cb = plt.colorbar(g)
            cb.set_label(hue) 
        elif plottype == 'seaborn':
            huedata = data.sort_values(by=hue, ascending=True)[hue].unique()
            im = sns.scatterplot(x=geo_x, y=geo_y, hue=hue, data=data, alpha=alpha, legend='brief',palette=palette, edgecolor='silver', linewidth=0.5,hue_order=huedata)
            # https://stackoverflow.com/questions/53437462/how-do-i-remove-an-attribute-from-the-legend-of-a-scatter-plot
            # EXTRACT CURRENT HANDLES AND LABELS
            h, l = self.ax.get_legend_handles_labels()
            # COLOR LEGEND (FIRST guess at size ITEMS) we don;t want to plot the distanceinc
            huelen = len(data.sort_values(by=hue, ascending=True)[hue].unique()) + 1
            col_lgd = plt.legend(h[:huelen], l[:huelen], loc='upper left', bbox_to_anchor=(1.05, 1), fancybox=True, shadow=True, ncol=1)
            plt.gca().add_artist(col_lgd)
            self.ax.set_xlabel('')
            self.ax.set_ylabel('')
            plt.legend(title=hue,bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)  # Put the legend out of the figure
        elif plottype == 'barplot':
            if palette == "viridis":
                palette = 'b'            
            self.ax.bar(data[geo_x],data[geo_y],color=palette,width=0.5)        
            self.ax.set_xlabel('')
            self.ax.set_ylabel('')            
        elif plottype == 'hist2d':
            x = data[geo_x]
            y = data[geo_y]
            bins=100
            gridsize=50
            hb = plt.hexbin(x, y, bins=bins, cmap=palette, gridsize=gridsize,extent=[minx,maxx,miny,maxy])
        elif plottype == 'probability':
            contours = 12
            bins = 50
            plt.axis([minx, maxx, miny, maxy])
            xgrid = np.linspace(minx, maxx, bins)
            ygrid = np.linspace(miny, maxy, bins)
            xdata = data[geo_x]
            ydata = data[geo_y]
            data = np.vstack([xdata, ydata])
            xgrid = np.linspace(minx, maxx, bins)
            ygrid = np.linspace(miny, maxy, bins)
            Xgrid, Ygrid = np.meshgrid(xgrid, ygrid)
            grid_sized = np.vstack([Xgrid.ravel(), Ygrid.ravel()])
            # fit an array of size [Ndim, Nsamples]
            kde = gaussian_kde(data, bw_method=0.1)
            # evaluate on a regular grid
            Z = kde.evaluate(grid_sized)
            zgrid = Z.reshape(Xgrid.shape)
            self.ax.grid(True, which='major', axis='both', linestyle='-', color=(0.5, 0.5, 0.5), alpha=0.1)
            im = plt.pcolormesh(xgrid, ygrid, zgrid, shading='gouraud', cmap=palette, alpha=alpha)
            cs = plt.contour(xgrid, ygrid, zgrid, contours, colors='0.7', linewidths=0.4, alpha=alpha,extent=[minx,maxx,miny,maxy])
            self.ax.set_axisbelow(True)

        self.ax.set_xlabel(geo_x + "\nCount=" + str(count))
        self.ax.set_ylabel(geo_y)
        plt.title(title)
        plt.xlim([minx, maxx])
        plt.ylim([miny, maxy])
        plt.xticks(rotation=45)   
        #Having plotted we now need to get the image data from the plt
        if not overlay:
            encoded = self.getPlotImage(self.fig, self.ax)
            htmlstring = '<img src="data:image/png;base64, {}">'.format(encoded.decode('utf-8')) + '\n'
            self.HTMLIncrement()
            self.html_string += '<td width=' + str(int(100/self.cols)) + '%>' + htmlstring + '</td>\n'

    def addPoints2d(self,points,hue="yellow",overlay=False):
        self.incrementOverlay(overlay)        
        self.ax.set_axisbelow(True)
        xs = []
        ys = []
        for p in points:
            xs.append(p[0])
            ys.append(p[1])
        g = self.ax.scatter(xs, ys, edgecolor='silver', linewidth=0.5, s=20,c=hue)
        cb = plt.colorbar(g)
        cb.remove() 
        #Having plotted we now need to get the image data from the plt
        if not overlay:
            encoded = self.getPlotImage(self.fig, self.ax)
            htmlstring = '<img src="data:image/png;base64, {}">'.format(encoded.decode('utf-8')) + '\n'
            self.HTMLIncrement()
            self.html_string += '<td width=' + str(int(100/self.cols)) + '%>' + htmlstring + '</td>\n'
        
        
        

    def addPlotOnly(self,fig,ax):
        encoded = self.getPlotImage(fig,ax)
        htmlstring = '<img src="data:image/png;base64, {}">'.format(encoded.decode('utf-8')) + '\n'
        self.HTMLIncrement()
        self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>' + htmlstring + '</td>\n'

    def addPlot1d(self, data,plottype, geo_x,hue='',title='',palette='crimson',overlay=False,alpha=1,xrange=[None,None],cumulative=False,density=False,bins=20):
        self.incrementOverlay(overlay)
        axisrange = [xrange[0],xrange[1]] # by ref obs need to be copied
        if xrange[0] == None:
            axisrange[0] = min(data[geo_x])
        if xrange[1] == None:
            axisrange[1] = max(data[geo_x])

        if len(geo_x[0]) > 1: #todo evident bug if the x is a 1 lne stingf
            mydata = data[geo_x]
            mydata = mydata.transpose()
        else:
            mydata = data[geo_x]

        if plottype == 'histogram':
            g = plt.hist(mydata, edgecolor='k', bins=bins, color=palette, alpha=alpha, label='geo_x',range=axisrange,cumulative=cumulative,density=density)
        elif plottype == 'step':
            g = plt.hist(mydata, edgecolor='k', bins=bins, color=palette, alpha=alpha, label='geo_x',range=axisrange,cumulative=cumulative,density=density,histtype='step')
        elif plottype == 'stepfilled':
            g = plt.hist(mydata, edgecolor='k', bins=bins, color=palette, alpha=alpha, label='geo_x',range=axisrange,cumulative=cumulative,density=density,histtype='stepfilled')
        elif plottype == 'barstacked':
            g = plt.hist(mydata, edgecolor='k', bins=bins, color=palette, alpha=alpha, label='geo_x',range=axisrange,cumulative=cumulative,density=density,histtype='barstacked')
        elif plottype == 'violin':
            g = plt.violinplot(mydata,showmeans=False,showextrema=True,showmedians=True,quantiles=[0.25,0.75])
        
        #cb = plt.colorbar(g)
        #hue is used to find the outliers
        if hue != '':
            geos = []
            if len(geo_x[0]) > 1:  # todo evident bug if the x is a 1 lne stingf
                geos = geo_x
            else:
                geos = [geo_x]

            for geo in geos:
                data = data.sort_values(by=geo,ascending=True)
                firstval = data.head(1)[hue].values[0]
                try:
                    firstval = str(round(firstval))
                except:
                    pass
                firstgeo = round(data.head(1)[geo].values[0],3)
                data = data.sort_values(by=geo, ascending=False)
                lastval = data.head(1)[hue].values[0]
                try:
                    lastval = str(round(lastval))
                except:
                    pass
                lastgeo = round(data.head(1)[geo].values[0],3)
                title += '\n' + geo + ' ' + hue + ': ' + str(firstval)+ '=' + str(firstgeo) + ' ' + str(lastval) + '=' + str(lastgeo)

        count = len(data[geo_x])
        self.ax.set_xlabel(geo_x + "\nCount=" + str(count))
        plt.title(title)
        # Having plotted we now need to get the image data from the plt
        if not overlay:
            encoded = self.getPlotImage(self.fig, self.ax)
            htmlstring = '<img src="data:image/png;base64, {}">'.format(encoded.decode('utf-8')) + '\n'
            self.HTMLIncrement()
            self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>' + htmlstring + '</td>\n'

    def addPlotPi(self, data,geo_x,hue,title='',colors = [],overlay=False,percent=False):
        self.incrementOverlay(overlay)
        if colors == []:
            if percent:
                plt.pie(data[geo_x],labels=data[hue], autopct='%1.1f%%')
            else:
                plt.pie(data[geo_x],labels=data[hue])
        else:
            if percent:
                plt.pie(data[geo_x], labels=data[hue],colors=colors,autopct='%1.1f%%')
            else:
                plt.pie(data[geo_x], labels=data[hue],colors=colors)        
        plt.title(title)
        # Having plotted we now need to get the image data from the plt
        if not overlay:
            encoded = self.getPlotImage(self.fig, self.ax)
            htmlstring = '<img src="data:image/png;base64, {}">'.format(encoded.decode('utf-8')) + '\n'
            self.HTMLIncrement()
            self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>' + htmlstring + '</td>\n'

    def addSeries(self,data,title='',transpose=False):
        if transpose:
            self.addSeries_Transposed(data,title)
        else:
            self.addSeries_Not(data,title)

    def addSeries_Transposed(self, data, title=''):
        rows = len(data.index)
        idx_names = list(data.index)
        innerhtml = "<table>\n"
        innerhtml += "<tr>\n"
        for r in range(0, rows):
            innerhtml += "<td>" + str(idx_names[r]) + "</td>\n"
        innerhtml += "</tr>\n"
        innerhtml += "<tr>"
        for r in range(0, rows):
            header = idx_names[r]
            innerhtml += "<td>"
            try:#make a guess at how to round it if it is a number
                number = float(data[r])
                strnumber = str(round(number, 1))
                if abs(number) < 10:
                    strnumber = str(round(number, 3))
                elif abs(number) < 100:
                    strnumber = str(round(number, 2))
                elif abs(number) > 1000:
                    strnumber = str(int(round(number, 0)))
                # print(header, number,strnumber)
                innerhtml += strnumber

            except:
                innerhtml += str(data[r])
            innerhtml += "</td>\n"

        innerhtml += "</tr>\n"
        innerhtml += "</table>\n"

        self.HTMLIncrement()
        self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>' + title + '</br>' + innerhtml + '</td>\n'

    def addSeries_Not(self, data, title=''):
        rows = len(data.index)
        idx_names = list(data.index)
        innerhtml = "<table>\n"
        for r in range(0, rows):
            innerhtml += "<tr><td>" + str(idx_names[r]) + "</td>\n"
            header = idx_names[r]
            innerhtml += "<td>"
            try:#make a guess at how to round it if it is a number
                number = float(data[r])
                strnumber = str(round(number, 1))
                if abs(number) < 10:
                    strnumber = str(round(number, 3))
                elif abs(number) < 100:
                    strnumber = str(round(number, 2))
                elif abs(number) > 1000:
                    strnumber = str(int(round(number, 0)))
                # print(header, number,strnumber)
                innerhtml += strnumber
            except:
                innerhtml += str(data[r])
            innerhtml += "</td>\n"
            innerhtml += "</tr>\n"
        innerhtml += "</table>\n"

        self.HTMLIncrement()
        self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>' + title + '</br>' + innerhtml + '</td>\n'

    def mtxCap(self,mtx,cap):
        if cap < 0:
            return self.mtxCapMin(mtx,cap)
        else:
            mtxret = np.zeros(mtx.shape)
            mmin,mmax = np.amin(mtx),np.max(mtx)
            capped = mmin + cap*(mmax-mmin)
            a,b = mtx.shape
            for i in range(a):
                for j in range(b):
                    if mtx[i,j] > capped:
                        mtxret[i, j] = capped
                    else:
                        mtxret[i, j] = mtx[i, j]
            return mtxret

    def mtxCapMin(self,mtx,cap):
        mtxret = np.zeros(mtx.shape)
        mmin,mmax = np.amin(mtx),np.max(mtx)
        capped = mmax + cap*(mmax-mmin)
        a,b = mtx.shape
        for i in range(a):
            for j in range(b):
                if mtx[i,j] < capped:
                    mtxret[i, j] = capped
                else:
                    mtxret[i, j] = mtx[i, j]
        return mtxret


    def incrementOverlay(self,overlay):
        if overlay:
            if not self.overlay_open:
                self.fig, self.ax = plt.subplots()
                self.overlay_open = True
        else:
            if not self.overlay_open:
                self.fig, self.ax = plt.subplots()
            self.overlay_open = False

    def addSurface(self, mtx, title='',palette='inferno',alpha=0.9,overlay=False,cmin=None,cmax=None,cap=0,colourbar=False,centred=False):
        self.incrementOverlay(overlay)

        if cap != 0:
            mtx = self.mtxCap(mtx,cap)

        vmin, vmax = np.amin(mtx), np.max(mtx)
        if cmin != None:
            vmin = cmin
        if cmax != None:
            vmax = cmax

        image = plt.imshow(mtx, cmap=palette, interpolation='nearest', origin='lower', aspect='equal', alpha=alpha,vmin=vmin, vmax=vmax)

        plt.axis('off')
        plt.title(title)

        if colourbar:
            self.fig.colorbar(image, ax=self.ax)
        # Having plotted we now need to get the image data from the plt
        if not overlay:
            encoded = self.getPlotImage(self.fig, self.ax)
            htmlstring = '<img src="data:image/png;base64, {}">'.format(encoded.decode('utf-8')) + '\n'
            self.HTMLIncrement()
            self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>' + htmlstring + '</td>\n'

    def addContours(self, mtx, title='',style='filled', contourlabel=False,levels=10,centred=False,palette='inferno',alpha=0.9,colourbar=True,overlay=False,cmin=None,cmax=None,cap=0,linewidth=None):
        '''
        :param mtx:
        :param title:
        :param style: Can be filled, lines or both
        :param label: label the contours
        :param levels: can be number or cohosen list
        :param centres: sets 0 as the middle in the colours (use divergent colour map)
        :param palette:
        :param alpha:
        :param overlay:
        :return:
        '''
        self.incrementOverlay(overlay)

        # Need to turn the matrix into something countourable
        if cap != 0:
            mtx = self.mtxCap(mtx,cap)

        vmin,vmax = np.amin(mtx),np.max(mtx)
        if cmin != None:
            vmin = cmin
        if cmax != None:
            vmax = cmax
        elif centred:
            vmax = max(abs(vmin),abs(vmax))
            vmin = -1 * vmax

        #Need the X and Z
        abc = mtx.shape
        a,b = abc[0],abc[1]
        x = np.arange(0,a,1)
        y = np.arange(0, b, 1)
        X,Y = np.meshgrid(x,y)
        Z = mtx

        self.ax.set_aspect('equal')
        self.ax.axes.xaxis.set_ticklabels([])
        self.ax.axes.yaxis.set_ticklabels([])

        if style=='filled':
            cf = self.ax.contourf(X,Y,Z,cmap=palette,levels=levels,alpha=alpha,vmin=vmin,vmax=vmax,extend='both')
        elif style == 'lines':
            cf = self.ax.contour(X, Y, Z, cmap=palette,levels=levels, alpha=alpha,vmin=vmin,vmax=vmax,extend='both',linewidths=linewidth)
        else:
            cf = self.ax.contourf(X, Y, Z, cmap=palette,levels=levels,alpha=alpha,vmin=vmin,vmax=vmax,extend='both')
            cf2 = self.ax.contour(X, Y, Z, colors='black',levels=levels,vmin=vmin,vmax=vmax,extend='both',linewidths=linewidth)

        if contourlabel:
            if style=='both':
                self.ax.clabel(cf2,inline=True,fontsize=8)
            elif style=='lines':
                self.ax.clabel(cf, inline=True, fontsize=8)
            # can't put contours names without lines

        if colourbar:
            self.fig.colorbar(cf,ax=self.ax)

        plt.axis('off')
        plt.title(title)
        # Having plotted we now need to get the image data from the plt
        if not overlay:
            encoded = self.getPlotImage(self.fig, self.ax)
            htmlstring = '<img src="data:image/png;base64, {}">'.format(encoded.decode('utf-8')) + '\n'
            self.HTMLIncrement()
            self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>' + htmlstring + '</td>\n'

    def addDataFrame(self, data, title=''):
        rows = len(data.index)
        cols = data.columns
        innerhtml = "<table>\n"
        innerhtml += "<tr>\n"
        for col in cols:
            innerhtml += "<td>" + str(col) + "</td>\n"
        innerhtml += "</tr>\n"

        for r in range(0, rows):
            innerhtml += "<tr>"
            for col in cols:
                innerhtml += "<td>"
                collist = data[col].tolist()
                strval = str(collist[r])
                innerhtml += strval
                innerhtml += "</td>\n"

        innerhtml += "</tr>\n"
        innerhtml += "</table>\n"

        self.HTMLIncrement()
        self.html_string += '<td width=' + str(int(100 / self.cols)) + '%>' + title + '</br>' + innerhtml + '</td>\n'

    def HTMLIncrement(self):
        if not self.table_open:
            self.html_string += '<table><tr>\n'
            self.table_open = True
        if self.num_col == self.cols:
            self.html_string += '</tr><tr>\n'
            self.num_col = 0
        self.num_col += 1

    def getPlotImage(self,fig, ax):
        img = io.BytesIO()
        fig.savefig(img, format='png', bbox_inches='tight')
        img.seek(0)
        encoded = base64.b64encode(img.getvalue())
        plt.close('all')
        return encoded

    def getHeaderString(self,title, remove_strip):
        html = '<!DOCTYPE html><html lang="en"><head><title>LeucipPy Report</title>\n'
        # html += '<style> body {background-color:SeaShell;} table {table-layout:fixed;display:table;margin:0 auto;}td {border:1px solid RosyBrown;background-color:SeaShell;}</style>'
        # html += '<style> body {background-color:HoneyDew;} table {background-color:HoneyDew;} .innertable td {border:1px solid MistyRose;background-color:MintCream;}</style>'
        html += '<style> body {text-align:center;background-color:' + self.clr_background + ' ;} img {width:95% }'
        html += 'table {font-size:0.8vw;width:95%;table-layout:fixed;display:table;margin:0 auto;background-color:lightgrey ;}'
        html += ' td {border:1px solid ' + self.clr_behindplot + ';background-color:' + self.clr_behindplot + ';}</style>'
        html += '</head>\n'
        html += '<body>'
        # if not remove_strip:
        #    html += '<hr/>'
        #    html += '<div style="background-color:' + self.clr_header + ';padding:10px"> LeucipPy: Protein Geometry Correlations </div>'
        html += '<hr/>'
        html += '<div style="background-color:lightgrey;padding:5px">'
        html += '<h1>' + title + '</h1>\n'
        html += '</div>'
        html += '<hr/>\n'

        return html

    def getFooterString(self):
        html = '<hr/><div style = "background-color:' + self.clr_header +';padding:10px" >'
        html += '<a href = "https://rachelalcraft.github.io/leucipPy.html" title = "LeucipPy" target = "_self">LeucipPy</a>'
        html += ' by <a href = "mailto:rachelalcraft@gmail.com">Rachel Alcraft</a>'
        html += ' ~ supervisor <a href = "http://people.cryst.bbk.ac.uk/~ubcg66a/">Mark A Williams</a>'
        html += ' ~ Birkbeck, University of London (2021)'
        html += ' ~ Follow <a href = "https://rachelalcraft.github.io/citing.html"> citation guidance </a> </br></div><hr/>'
        return html





"""
RSA 4/2/23
https://pynative.com/make-python-class-json-serializable/#:~:text=Use%20toJSON()%20Method%20to%20make%20class%20JSON%20serializable&text=So%20we%20don't%20need,Python%20Object%20to%20JSON%20string.

"""

from maptial.xyz import vectorthree as v3
import pandas as pd
import json

amino_acids = ["ala","arg","asn","asp","cys","gln","glu","gly","his","ile","leu","lys","met","phe","pro","ser","thr","trp","tyr","val"]
class PdbObject(object):
    def __init__(self, pdb_code):
        # PUBLIC INTERFACE        
        self.pdb_code = pdb_code
        self.resolution = -1
        self.exp_method = ""
        self.chains = {}
        self.exc_hetatm = False
        self.lines = []
              
    def __str__(self):
        return f"{self.pdb_code}\t{self.resolution}\t{self.exp_method}"
    
    def lines(self,atoms=False):
        lines = []
        for chain,resdic in self.chains.items():
            for no,res in resdic.items():                
                lines.append(chain + "\t" + str(res))
                for attype,atm in res.atoms.items():
                    if atoms:
                        lines.append(atm)
        return lines
                                
    def add_atoms(self,bio_struc):
        self.bio_struc = bio_struc
        self.resolution = self.bio_struc.header['resolution']        
        self.exp_method = self.bio_struc.header["structure_method"]                
        atomNo = 0
        ridx = 0
        last_bad = ""
        for model in self.bio_struc:
            for chain in model:
                self.chains[chain.id] = {} ####  a chain is a dictionary of residue number to GeoResidue ############################################################
                for residue in chain:
                    aa = residue.get_resname()
                    rid = residue.get_full_id()[3][1]                    
                    resd = PdbResidue(aa,rid,ridx)
                    chain = residue.get_full_id()[2]                    
                    hetatm = residue.get_full_id()[3][0]
                    aah = residue.get_full_id()
                    mutation = aah[3][2] #these are alternative conformations where mutations could occur do not add
                    # in fact remove any that area lready there
                    if mutation.strip() != "":
                        #self.chains[chain][rid] = resd
                        if chain in self.chains:
                            if rid in self.chains[chain]:
                                self.chains[chain].pop(rid)
                    elif mutation.strip() == "":
                        this_bad = chain + "_" + str(rid)
                        if this_bad != last_bad:                                    
                            if rid in self.chains[chain] and self.exc_hetatm:                          
                                # if it is already there something is wrong so lets remove it
                                del self.chains[chain][rid]
                            elif mutation[0] != " " and self.exc_hetatm:                          
                                # if there are mutation insertions they cause us geoemtry problems and we will remove them                        
                                last_bad = chain + "_" + str(rid)
                                if rid in self.chains[chain]: #we know this has alreadybeen checked, it is just a line of code to stop it adding
                                    del self.chains[chain][rid]
                            else:                        
                                if str(hetatm[0:2]) == "H_" and self.exc_hetatm:
                                    if False:
                                        print("HETATM",rid)
                                else:
                                    # only proceed if we explicitly want hetatms                        
                                    ridx = ridx + 1

                                    for atom in residue:
                                        disordered = 'N'
                                        if atom.is_disordered():
                                            disordered = 'Y'
                                            if atom.disordered_has_id("A"):
                                                atom.disordered_select("A")
                                        if atom.get_occupancy() == None:
                                            disordered = 'Y'
                                        elif atom.get_occupancy() < 1:
                                            disordered = 'Y'
                                        atomNo += 1
                                        atom_name = atom.get_name()
                                        occupant = atom.get_full_id()[4][1]
                                        if occupant == ' ':
                                            occupant = 'A'
                                        x = atom.get_vector()[0]
                                        y = atom.get_vector()[1]
                                        z = atom.get_vector()[2]
                                        bfactor = atom.get_bfactor()
                                        occupancy = atom.get_occupancy()
                                        one_atom = PdbAtom(chain,resd,atom_name[0],atom_name,atomNo,disordered,occupancy,bfactor,x,y,z)
                                        self.add_line_string(atomNo, rid, aa, atom_name, chain, occupant, x, y, z, occupancy, bfactor, atom_name[0])
                                        resd.atoms[atom_name] = one_atom
                                    self.chains[chain][rid] = resd
                                    last_bad = ""

    def elementInList(self,element, atomlist):
        for atm in atomlist:
            if element in atm:
                return True, atm
        return False, None
    
    ######## MAP interface #####################
    def get_first_three(self):
        atm1, atm2, atm3 = {},{},{}
        count = 0
        for atm in self.lines:
            if atm["atm"] == "CA" and (atm["version"] == "A" or atm["version"] == ""):
                atm1 = atm
                count += 1
            elif atm["atm"] == "C" and (atm["version"] == "A" or atm["version"] == ""):
                atm2 = atm
                count += 1
            elif atm["atm"] == "O" and (atm["version"] == "A" or atm["version"] == ""):
                atm3 = atm
                count += 1
            if count == 3:
                return atm1,atm2,atm3
        return atm1,atm2,atm3
    
    def get_coords_key(self,key):
        atm = self.get_atm_key(key)
        if atm == {} or atm == None:
            return "(0,0,0)"
        return f"({atm['x']},{atm['y']},{atm['z']})"

    def get_coords(self,atm):
        if atm == {}:
            return "(0,0,0)"
        return f"({atm['x']},{atm['y']},{atm['z']})"
    
    def get_atm_key(self,key):
        #A:709@C.A
        try:
            sx = key.split("@")
            s01 = sx[0].split(":")
            s23 = sx[1].split(".")
            ch,rid,am,ver = s01[0],s01[1],s23[0],s23[1]            
            for atm in self.lines:                                
                if atm["chain"] == ch:
                    if int(atm["rid"]) == int(rid):
                        if atm["atm"] == am:
                            if atm["version"] == ver: 
                                return atm
        except Exception as e:
            print("Error finding key",key,e)
            return {}
        
    def get_key(self,atm):
        if atm == {}:
            return ""
        #A:709@C.A
        return f"{atm['chain']}:{atm['rid']}@{atm['atm']}.{atm['version']}"
    
    def get_description(self,atm,dis):
        if atm == {}:
            return ""
        #A:709@C.A
        return f"{round(dis,3)}:{atm['aa']}:{atm['chain']}:{atm['rid']}@{atm['atm']}.{atm['version']}"
    
    def get_next_key(self,key, offset=1):
        try:
            if len(key) < 4:
                return ""
            atm = self.get_atm_key(key)
            ridn = int(atm["rid"]) + offset
            return f"{atm['chain']}:{ridn}@{atm['atm']}.{atm['version']}"
        except:
            return ""

    def get_atom_coords(self):
        atoms = []
        for atm in self.lines:
            coord = v3.VectorThree(atm["x"],atm["y"],atm["z"])
            atoms.append(coord)
        return atoms
    
    def get_inscope_atoms(self,anchor,distance,log_level=0):
        in_scope = []
        for atm in self.lines:
            crd = v3.VectorThree(float(atm["x"]),float(atm["y"]),float(atm["z"]))
            dis = anchor.distance(crd)
            if dis <= distance:
                in_scope.append(atm)
                if log_level > 0:                        
                    print("neighbour in scope", self.get_description(atm,dis))
        return in_scope
    
    def get_neighbours(self,coord,rnge,atoms=None,ishtml=False):
        a = ""
        if atoms == None:
            atoms = self.lines
        for atm in atoms:
            crd = v3.VectorThree(float(atm["x"]),float(atm["y"]),float(atm["z"]))
            dis = coord.distance(crd)
            if dis >= rnge[0] and dis <= rnge[1]:
                desc = self.get_description(atm,dis)
                if desc not in a:
                    if ishtml:
                        a+="<br>......"+ desc
                    else:
                        a+="\n......"+desc
        return a
    ##################################################################################
    def dataFrame(self):        
        dicdfs = []        
        for chain,resdic in self.chains.items():
            for no,res in resdic.items():
                for attype,atm in res.atoms.items():     
                    dic={'pdbCode':self.pdb_code,'resolution':self.resolution,
                    'chain':atm.chain,'aa':res.amino_acid,'rid':res.rid,'ridx':res.ridx,                    
                    'atom':atm.atom_name, 'atomNo':atm.atom_no,'element':atm.atom_type,
                    'bfactor':atm.bfactor, 'occupancy':atm.occupancy,
                    'x':atm.x, 'y':atm.y, 'z':atm.z}
                    dicdfs.append(dic)
        return pd.DataFrame.from_dict(dicdfs)
    
    # if we add lines from a cif file I will do something different    
    def add_line_string(self,aid, rid, aa, am, ch, ver, x, y, z, occ, bf, ele):
        #ATOM    435  OE1AGLU A  23      12.418  13.330   0.541  0.58  7.57           O  
        #if "ATOM" in line[:6] or "HETATM" in line[:8]:                        
            #aid = line[6:12].strip()
            #am = line[12:16].strip()
            #ver = line[16:17].strip()
            #aa = line[17:21].strip()
            #ch = line[21:23].strip()
            #rid = line[23:29].strip()
            #x = line[29:38].strip()
            #y = line[38:46].strip()
            #z = line[46:54].strip()
            #occ = line[54:60].strip()
            #bf = line[60:66].strip()
            #ele = line[66:].strip()
        atm = {}
        atm["rid"] = rid
        atm["aid"] = aid
        atm["aa"] = aa
        atm["atm"] = am
        atm["chain"] = ch
        if ver == "":
            ver = "A"
        atm["version"] = ver            
        atm["x"] = x
        atm["y"] = y
        atm["z"] = z
        atm["occupancy"] = occ
        atm["bfactor"] = bf
        atm["element"] = ele        
        self.lines.append(atm)
        #if "HETATM" in line[:8]:
        #    print(line)
        #    print(atm)

    #################################
    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)
    
    def fromJson(self, jsndic):    
        self.pdb_code = jsndic["pdb_code"]
        self.lines = jsndic["lines"]


    
###################################################################################################################
class PdbResidue:
    """Class for a single residue

    :data amino acid:
    :data atoms:
    """

    def __init__(self, amino_acid,rid,ridx):
        """Initialises a GeoPdb with a biopython structure

        :param biopython_structure: A list of structures that has been created from biopython
        """

        self.atoms = {}
        self.elements = {}
        self.amino_acid = amino_acid
        self.rid = rid
        self.ridx = ridx

    def __str__(self):
       delim = "\t"
       return f"{self.amino_acid}{delim}{self.rid}{delim}{self.ridx}"
    
    def infoResd(self):
       delim = "|"
       return f"{self.amino_acid}{delim}{self.rid}"

###################################################################################################################
class PdbAtom:
    """Class for a single atom

    :data atom_type:
    :data atom_name:
    :data x:
    :data y:
    :data z:
    :data disordered:
    :data occupancy:
    :data bfactor:
    """

    def __init__(self, chain,res,atom_type,atom_name,atom_no,disordered,occupancy,bfactor,x,y,z):
        """Initialises a pdb with a biopython structure

        :param biopython_structure: A list of structures that has been created from biopython
        """

        self.info = {}
        self.chain = chain
        self.res = res
        self.atom_type = atom_type
        self.atom_name = atom_name
        self.atom_no = atom_no
        self.disordered = disordered
        self.occupancy = occupancy
        self.bfactor = bfactor
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        delim = "\t"
        return f"{self.atom_type}{delim}{self.atom_name}{delim}{self.atom_no}{delim}{self.disordered}{delim}{self.occupancy}{delim}{self.bfactor}{delim}({self.x}{self.y}{self.z})"

    def infoAtom(self):
        delim = "|"
        return f"({self.chain}{delim}{self.res.infoResd()}{delim}{self.atom_name}{delim}{self.atom_no})"

    def matchesCriteria(self, criteria,rids=[],dis=-1):        
        if self.disordered == "Y":
            return False
        if criteria == "":
            return True
        crits = criteria.split(",")
        for crit in crits:
            cri = crit.split("|")
            if cri[0].lower() == "aa":
                if self.res.amino_acid.upper() == cri[1].upper() or (cri[1].upper()=="20" and self.res.amino_acid.lower() in amino_acids) :
                    pass
                else:
                    return False
            if cri[0].lower() == "~aa":
                if self.res.amino_acid.upper() == cri[1].upper():
                    return False
                else:
                    pass
            
            elif cri[0].lower() == "rid" and len(rids) > 0:
                if "><" in cri[1]: # this is between the values
                    crits = cri[1].split('>')
                    crit_dis1 = float(crits[0])
                    crit_dis2 = float(crits[1][1:])
                    for rid in rids:
                        if abs(int(rid)-int(self.res.rid)) >= int(crit_dis1) and abs(int(rid)-int(self.res.rid)) <= int(crit_dis2):
                            pass
                        else:
                            return False
                elif "<>" in cri[1]: # these are at the extremes
                    crits = cri[1].split('<')
                    crit_dis1 = float(crits[0])
                    crit_dis2 = float(crits[1][1:])
                    for rid in rids:
                        if abs(int(rid)-int(self.res.rid)) <= int(crit_dis1) or abs(int(rid)-int(self.res.rid)) >= int(crit_dis2):
                            pass
                        else:
                            return False
                elif ">" in cri[1]:
                    crit_dis = cri[1][1:]
                    for rid in rids:
                        if abs(int(rid)-int(self.res.rid)) >= int(crit_dis):
                            pass
                        else:
                            return False
                elif "<" in cri[1]:
                    crit_dis = cri[1][1:]
                    for rid in rids:
                        if abs(int(rid)-int(self.res.rid)) <= int(crit_dis):
                            pass
                        else:
                            return False

            elif cri[0].lower() == "dis" and dis != -1:
                if "><" in cri[1]: #betweem
                    crits = cri[1].split('>')
                    crit_dis1 = float(crits[0])
                    crit_dis2 = float(crits[1][1:])
                    if dis >= crit_dis1 and dis <= crit_dis2:
                        pass
                    else:
                        return False
                elif "<>" in cri[1]: #extremes
                    crits = cri[1].split('<')
                    crit_dis1 = float(crits[0])
                    crit_dis2 = float(crits[1][1:])
                    if dis <= crit_dis1 or dis >= crit_dis2:
                        pass
                    else:
                        return False
                elif "<" in cri[1]:
                    crit_dis = cri[1][1:]
                    if float(dis) <= float(crit_dis):
                        pass
                    else:
                        return False            
                elif ">" in cri[1]:
                    crit_dis = cri[1][1:]
                    if float(dis) >= float(crit_dis):
                        pass
                    else:
                        return False
            
            elif cri[0].lower() == "occ":
                if "=" in cri[1]:
                    crit_occ = cri[1][1:]
                    if float(self.occupancy) == float(crit_occ):
                        pass
                    else:
                        return False            
                elif "<" in cri[1]:
                    crit_occ = cri[1][1:]
                    if float(self.occupancy) <= float(crit_occ):
                        pass
                    else:
                        return False            
                elif ">" in cri[1]:
                    crit_occ = cri[1][1:]
                    if float(self.occupancy) >= float(crit_occ):
                        pass
                    else:
                        return False
                
        return True
    



"""

RSA 25/2/23
This class handles transformations in 3d space from a plabe defined by 3 given points to the origin

"""

import math
from maptial.xyz import vectorthree as v3
from maptial.xyz import matrix3d as d3

class SpaceTransform(object):
    def __init__(self, central, linear, planar, log=False):
        # PUBLIC INTERFACE
        self.central = central
        self.linear = linear
        self.planar = planar

        #There are a set of transformations that need to be done in 1 order or the other to convert to the origin and then away.
        self._1_translation = v3.VectorThree(0,0,0)
        self._2_rotationXY = 0.0
        self._3_rotationXZ = 0.0
        self._4_rotationYZ = 0.0        
        #orthogonal unit axes       
        self.xAxis = v3.VectorThree(0,0,0)
        self.yAxis = v3.VectorThree(0,0,0)
        self.zAxis = v3.VectorThree(0,0,0)
        self.xOrthog = v3.VectorThree(0,0,0)
        self.yOrthog = v3.VectorThree(0,0,0)
        self.zOrthog = v3.VectorThree(0,0,0)
        self.centre = v3.VectorThree(0,0,0)

        self.M_PI = 3.14159265358979323846;

        self._calculate_transformation(log)

    
    ##### INTERNAL INTERFACE #######
    def _calculate_transformation(self,log):
        #First transformation is to the origin, a translation
        self._1_translation = self.central
        lin = self.linear
        pla = self.planar
        lin = v3.v3_subtract(lin,self._1_translation)
        pla = v3.v3_subtract(pla,self._1_translation)
        
        #Rotation vectors                    
        vR = v3.VectorThree(0,0,0)

        #Second transformation is to rotate the linear vector to make y=0                        
        self._2_rotationXY = self.get_rotation_angle(lin.A, lin.B)
        vR = self.rotate(lin.A, lin.B, self._2_rotationXY)
        lin.A = vR.A
        lin.B = vR.B
        vR = self.rotate(pla.A, pla.B, self._2_rotationXY)
        pla.A = vR.A
        pla.B = vR.B

        #Third transformation is to rotate the linear vector to make z=0            
        self._3_rotationXZ = self.get_rotation_angle(lin.A, lin.C)
        vR = self.rotate(lin.A, lin.C, self._3_rotationXZ)
        lin.A = vR.A
        lin.C = vR.B
        vR = self.rotate(pla.A, pla.C, self._3_rotationXZ)
        pla.A = vR.A
        pla.C = vR.B

        #Fourth transformation is to rotate the planar vector to make z=0    
        self._4_rotationYZ = self.get_rotation_angle(pla.B, pla.C)
        vR = self.rotate(pla.A, pla.C, self._4_rotationYZ) 
        pla.B = vR.A
        pla.C = vR.B

        #We have the transformations, now set up the orthogonal axes
        #Centre (0,0,0) should go to the central point            
        self.centre = self.apply_transformation(v3.VectorThree(0, 0, 0))
        self.xOrthog = self.apply_transformation(v3.VectorThree(1, 0, 0))
        self.xAxis = v3.v3_subtract(self.xOrthog,self.centre)
        self.yOrthog = self.apply_transformation(v3.VectorThree(0, 1, 0))
        self.yAxis = v3.v3_subtract(self.yOrthog, self.centre)
        self.zOrthog = self.apply_transformation(v3.VectorThree(0, 0, 1))
        self.zAxis = v3.v3_subtract(self.zOrthog, self.centre)

        #If these are othogonal the dot products will be zero, this is a debugging check
        if log:
            ortho = v3.VectorThree(1, 0, 0)
            o0 = ortho.dot_product(v3.VectorThree(0, 1, 0))
            o1 = self.xAxis.dot_product(self.yAxis)
            o2 = self.xAxis.dot_product(self.zAxis)
            o3 = self.zAxis.dot_product(self.yAxis)
            print(o0, o1, o2, o3)

    def get_rotation_angle(self, x, y):        
        theta = 0.0
        qStart = self._get_quadrant(x, y)
        vA = v3.VectorThree(0,0,0)
        axis = v3.VectorThree(0,0,0)
        mag = math.sqrt(math.pow(x, 2) + math.pow(y, 2))
        if mag > 0.0001:
            vA.A = x
            vA.B = y
            axis.A = mag
            axis.B = 0
            theta = vA.get_angle(axis)
            #We want to go claockwise to the positive x-axis and this is just the absolute difference, so:        
            if qStart == 4 or qStart == 3:
                theta = 2 * self.M_PI - theta
            if theta < 0:
                theta = 2 * self.M_PI + theta
        
        return theta

    def rotate(self,x, y, angle):        
        angle_left = angle
        x_now = x
        y_now = y
        pointPrime = v3.VectorThree(0,0,0)
        while (angle_left > self.M_PI / 2):                    
            pointPrime = self._rotate_ninety(x_now, y_now)
            x_now = pointPrime.A
            y_now = pointPrime.B
            angle_left -= self.M_PI / 2        
        pointPrime = self._rotate_quadrant(x_now, y_now, angle_left)
        return pointPrime

    def apply_transformation(self, point):        
        pointPrime = v3.VectorThree(point.A, point.B, point.C)
        point2 = v3.VectorThree(0,0,0)

        rotationYZ_4 = self._4_rotationYZ;
        rotationXZ_3 = self._3_rotationXZ;
        rotationXY_2 = self._2_rotationXY;

        point2 = self.rotate(pointPrime.B, pointPrime.C, 2 * self.M_PI - rotationYZ_4)
        pointPrime.B = point2.A
        pointPrime.C = point2.B

        point2 = self.rotate(pointPrime.A, pointPrime.C, 2 * self.M_PI - rotationXZ_3)
        pointPrime.A = point2.A
        pointPrime.C = point2.B

        point2 = self.rotate(pointPrime.A, pointPrime.B, 2 * self.M_PI - rotationXY_2)
        pointPrime.A = point2.A
        pointPrime.B = point2.B

        pointPrime = v3.v3_add(pointPrime, self._1_translation)

        return pointPrime

    def reverse_transformation(self, point):        
        pointPrime = v3.VectorThree(point.A, point.B, point.C)
        point2 = v3.VectorThree(0,0,0)

        rotationYZ_4 = self._4_rotationYZ
        rotationXZ_3 = self._3_rotationXZ
        rotationXY_2 = self._2_rotationXY

        pointPrime = v3.v3_subtract(pointPrime, self._1_translation)

        point2 = self.rotate(pointPrime.A, pointPrime.B, rotationXY_2)
        pointPrime.A = point2.A
        pointPrime.B = point2.B

        point2 = self.rotate(pointPrime.A, pointPrime.C, rotationXZ_3)
        pointPrime.A = point2.A
        pointPrime.C = point2.B

        point2 = self.rotate(pointPrime.B, pointPrime.C, rotationYZ_4)
        pointPrime.B = point2.A
        pointPrime.C = point2.B

        return pointPrime

    def convert_coords(self,unit_coords):
        coords = []
        # 2d or 3d?
        if type(unit_coords) is d3.Matrix3d:
            return self.convert_coords3d(unit_coords)
        else:
            a,b = unit_coords.shape()
            mat2 = d3.Matrix3d(a,b)            
            for i in range(a):                
                for j in range(b):
                    vec = v3.VectorThree(unit_coords.get(i,j)[0],unit_coords.get(i,j)[1],0)
                    vec_t = self.apply_transformation(vec)                                    
                    mat2.add(i,j,vec_t)
            return mat2
    
    def convert_coords3d(self,unit_coords):        
        a,b,c = unit_coords.shape()
        mat3 = d3.Matrix3d(a,b,c)
        for i in range(a):            
            for j in range(b):                
                for k in range(c):
                    x = unit_coords.get(i,j,k=k)[0]
                    y = unit_coords.get(i,j,k=k)[1]
                    z = unit_coords.get(i,j,k=k)[2]
                    vec = v3.VectorThree(x,y,z)
                    vec_t = self.apply_transformation(vec)
                    mat3.add(i,j,k,vec_t)                            
        return mat3
                        
##### INTERNAL FUNCTIONS ###
    def _get_quadrant(self,x, y):                           
        qStart = 1
        #first if it is on an axis
        if (x == 0 and y > 0):
            qStart = 1
        elif (x == 0 and y < 0):
            qStart = 4
        elif (y == 0 and x > 0):
            qStart = 1
        elif (y == 0 and x < 0):
            qStart = 2
        elif (x < 0 and y < 0):
            qStart = 3
        elif (x < 0 and y > 0):
            qStart = 2
        elif (y < 0 and x > 0):
            qStart = 4
        return qStart
    
    def _rotate_ninety(self, x_now, y_now):
        q = self._get_quadrant(x_now, y_now)
        q -= 1
        if (q == 0):
            q = 4;
        nextQ = v3.VectorThree(abs(y_now), abs(x_now), 0)
        if (q == 2):
            nextQ.A *= -1
        elif (q == 3):        
            nextQ.A *= -1
            nextQ.B *= -1    
        elif (q == 4):   
            nextQ.B *= -1        
        return nextQ
        
    def _rotate_quadrant(self,x, y, angle):        
        if (abs(angle) > 0.001):        
            v = v3.VectorThree(0,0,0)
            #This assumes an angle that is positive and less or = than 90 that may turn only into the next quadrant.            
            #Choose quadrants
            qStart = self._get_quadrant(x, y)
            mag = math.sqrt(math.pow(x, 2) + math.pow(y, 2)) #the length of the vector
            if (mag > 0.0001):            
                sinA = abs(y) / mag;
                angleA = math.asin(sinA); #this is the angle made with the x-axis from the original vector
                angleB = angleA - angle; #this is the angle made with the x-axis with the rotated vector
                qEnd = qStart
                if (qStart == 1):                
                    if (angle > angleA):                    
                        angleB = angle - angleA
                        qEnd = 4                                    
                elif (qStart == 2):                
                    angleB = angle + angleA
                    if (angleA + angle > self.M_PI / 2):                    
                        angleB = self.M_PI - (angleA + angle)
                        qEnd = 1                                    
                elif (qStart == 3):                
                    if (angle > angleA):                    
                        angleB = angle - angleA
                        qEnd = 2                                    
                else: #must be q4                
                    angleB = angle + angleA
                    if (angleA + angle > self.M_PI / 2):                    
                        angleB = self.M_PI - (angle + angleA)
                        qEnd = 3                                    
                x2 = math.cos(angleB) * mag
                y2 = math.sin(angleB) * mag
                if (qEnd == 2 or qEnd == 3):
                    x2 *= -1
                if (qEnd == 3 or qEnd == 4):
                    y2 *= -1
                v.A = x2;#Math.Round(x2, 8);
                v.B = y2;#Math.Round(y2, 8);            
            return v        
        else:        
            return v3.VectorThree(x, y, 0)
    
    def navigate(self, point, nav, nav_distance,angle=-1):
        if angle < 0:
            angle = 2 * self.M_PI / 90
        point = self.reverse_transformation(point)
        if (nav == "DN"):#'DN'down
            point.A += nav_distance        
        elif (nav == "UP"):#'UP'up
            point.A -= nav_distance        
        elif (nav == "LE"):#'LE'left
            point.B += nav_distance        
        elif (nav == "RI"):#'RI'right
            point.B -= nav_distance        
        elif (nav == "FW"):#'FW'fwd
            point.C -= nav_distance        
        elif (nav == "BA"):#'BA'back
            point.C += nav_distance        
        elif (nav == "TL"):#'TL'tilt left       
            ll = self.rotate(point.B, point.C, angle)
            point.B = ll.A
            point.C = ll.B
        elif (nav == "TR"):#'TR'tilt right
            ll = self.rotate(point.B, point.C, -1 * angle)
            point.B = ll.A
            point.C = ll.B        
        elif (nav == "TO"):#'TO'tilt over
            ll = self.rotate(point.A, point.C, angle)
            point.A = ll.A
            point.C = ll.B        
        elif (nav == "TU"):#'TU'tilt under
            ll = self.rotate(point.A, point.C, -1 * angle)
            point.A = ll.A
            point.C = ll.B                                
        elif (nav == "CL"):#'CL'clockwise
            ll = self.rotate(point.A, point.B, angle)
            point.A = ll.A
            point.B = ll.B        
        elif (nav == "AC"):#'AC'anti-clockwise
            ll = self.rotate(point.A, point.B, -1*angle)
            point.A = ll.A
            point.B = ll.B        
        point = self.apply_transformation(point)
        return point
        
                        


"""

RSA 25/2/23
This class handles transformations in 3d space from a plabe defined by 3 given points to the origin

"""

from maptial.xyz import vectorthree as v3
from maptial.xyz import matrixthree as m3
from maptial.xyz import matrix3d as d3

import math

class CrsTransform(object):
    def __init__(self,  dim_order,
                        crs_starts,
                        axis_sampling,
                        map2crs,
                        cell_dims,
                        angles ):
        
        self.M_PI = 3.14159265358979323846
        
        # PUBLIC INTERFACE
        self.dim_order = dim_order
        self.crs_starts = crs_starts
        self.axis_sampling = axis_sampling        
        self.map2crs = map2crs
        self.cell_dims = cell_dims
        self.angles = angles        
        self.map2xyz = [0,0,0]
        self.map2xyz[self.map2crs[0]] = 0
        self.map2xyz[self.map2crs[1]] = 1
        self.map2xyz[self.map2crs[2]] = 2                                
        self._create_transformation()
        
    ########## PUBLIC INTERFACE #############
    def crs_to_xyz(self, CRS):
        vXYZ = v3.VectorThree()
        #If the axes are all orthogonal            
        if (self.angles[0] == 90 and self.angles[1] == 90 and self.angles[2] == 90):
            for i in range(3):
                startVal = CRS.get_by_idx(self.map2crs[i])
                startVal *= self.cell_dims[i] / self.axis_sampling[i]
                startVal += self.origin.get_by_idx(i)
                vXYZ.put_by_idx(i, startVal)
        else: # they are not orthogonal        
            vCRS = v3.VectorThree()
            for i in range(3):
                startVal = 0
                if (self.map2crs[0] == i):
                    startVal = self.crs_starts[0] + CRS.A
                elif (self.map2crs[1] == i):
                    startVal = self.crs_starts[1] + CRS.B
                else:
                    startVal = self.crs_starts[2] + CRS.C
                vCRS.put_by_idx(i, startVal);
            vCRS.put_by_idx(0, vCRS.get_by_idx(0) / self.axis_sampling[0])
            vCRS.put_by_idx(1, vCRS.get_by_idx(1) / self.axis_sampling[1])
            vCRS.put_by_idx(2, vCRS.get_by_idx(2) / self.axis_sampling[2])
            vXYZ = self.orthoMat.multiply(vCRS, False)        
        return vXYZ
        
    def xyz_to_crs(self, XYZ):                
        vCRS = v3.VectorThree()
        #If the axes are all orthogonal            
        if (self.angles[0] == 90 and self.angles[1] == 90 and self.angles[2]== 90):
            for i in range(3):            
                startVal = XYZ.get_by_idx(i) - self.origin.get_by_idx(i)
                startVal /= self.cell_dims[i] / self.axis_sampling[i]                
                vCRS.put_by_idx(i, startVal)        
        else: #they are not orthogonal        
            vFraction = self.deOrthoMat.multiply(XYZ, False)
            for i in range(3):            
                val = vFraction.get_by_idx(i) * self.axis_sampling[i] - self.crs_starts[self.map2xyz[i]];
                vCRS.put_by_idx(i, val);                    
        c = vCRS.get_by_idx(self.map2crs[0]);
        r = vCRS.get_by_idx(self.map2crs[1]);
        s = vCRS.get_by_idx(self.map2crs[2]);        
        return v3.VectorThree(c,r,s);

    def convert_coords_to_xyz(self,crs_coords):
        coords = []
        for i in range(len(crs_coords)):
            row = []
            for j in range(len(crs_coords[0])):                
                vec = crs_coords[i][j]
                vec_t = self.crs_to_xyz(vec)
                row.append(vec_t)
            coords.append(row)
        return coords
    
    def convert_coords_to_crs(self,xyz_coords):
        a,b,c = xyz_coords.shape()
        m3 = d3.Matrix3d(a,b,c)        
        for i in range(a):
            for j in range(b):
                for k in range(c):                        
                    vec = xyz_coords.get(i,j,k=k)
                    vec_t = self.xyz_to_crs(vec)
                    m3.add(i,j,k=k,data=vec_t)            
        return m3

    ########## PRIVATE INTERFACE #############
    def _make_ortho(self):        
        alpha = self.M_PI / 180 * self.angles[0]
        beta = self.M_PI / 180 * self.angles[1]
        gamma = self.M_PI / 180 * self.angles[2]
        temp = math.sqrt(1 - math.pow(math.cos(alpha), 2) - math.pow(math.cos(beta), 2) - math.pow(math.cos(gamma), 2) + 2 * math.cos(alpha) * math.cos(beta) * math.cos(gamma))
        v00 = self.cell_dims[0]
        v01 = self.cell_dims[1] * math.cos(gamma)
        v02 = self.cell_dims[2] * math.cos(beta)
        v10 = 0
        v11 = self.cell_dims[1] * math.sin(gamma)
        v12 = self.cell_dims[2] * (math.cos(alpha) - math.cos(beta) * math.cos(gamma)) / math.sin(gamma)
        v20 = 0
        v21 = 0
        v22 = self.cell_dims[2] * temp / math.sin(gamma)

        self.orthoMat = m3.MatrixThree([v00,v01,v02,v10,v11,v12,v20,v21,v22])
        #self.orthoMat = m3.MatrixThree([v00,v10,v20,v01,v11,v21,v02,v12,v22])
        #self.orthoMat = m3.MatrixThree()
        #self.orthoMat.put_value(v00, 0, 0)
        #self.orthoMat.put_value(v01, 0, 1)
        #self.orthoMat.put_value(v02, 0, 2)
        #self.orthoMat.put_value(v10, 1, 0)
        #self.orthoMat.put_value(v11, 1, 1)
        #self.orthoMat.put_value(v12, 1, 2)
        #self.orthoMat.put_value(v20, 2, 0)
        #self.orthoMat.put_value(v21, 2, 1)
        #self.orthoMat.put_value(v22, 2, 2)        
        self.deOrthoMat = self.orthoMat.get_inverse()

    def _make_origin(self):        
        oro = v3.VectorThree()
        for i in range(3):        
            startVal = 0;
            if (self.map2crs[0] == i):
                startVal = self.crs_starts[0]
            elif (self.map2crs[1] == i):
                startVal = self.crs_starts[1]
            else:
                startVal = self.crs_starts[2]
            oro.put_by_idx(i, startVal);        
        oro.put_by_idx(0, oro.get_by_idx(0) / self.axis_sampling[0])
        oro.put_by_idx(1, oro.get_by_idx(1) / self.axis_sampling[1])
        oro.put_by_idx(2, oro.get_by_idx(2) / self.axis_sampling[2])
        #self.origin = self.orthoMat.multiply(oro, True)
        self.origin = self.orthoMat.multiply(oro, False)
        
    def _create_transformation(self):        
        self._make_ortho()
        self._make_origin()               
        
        


"""

"""

import numpy as np
from numpy.linalg import inv
import math

class InvariantMaker(object):
    def __init__(self, dimensions):                
        self.round = 12
        dimX, dimY, dimZ = dimensions[0],1,1
        if len(dimensions) > 1:
            dimY = dimensions[1]
        if len(dimensions) > 2:
            dimZ = dimensions[2]

        simul = np.zeros([dimX*dimY*dimZ,dimX*dimY*dimZ])
        ser = -1
        for i in range(0, dimX):
            for j in range(0, dimY):
                for k in range(0, dimZ):
                    ser += 1
                    sec = -1
                    for ic in range(0, dimX):
                        for jc in range(0, dimY):
                            for kc in range(0, dimZ):
                                sec += 1
                                seCoeff = math.pow(i, ic) * math.pow(j, jc) * math.pow(k, kc)                                
                                simul[ser, sec] = seCoeff

        self.alcraft = np.round(inv(simul),14)

    def save_as_file(self,filename,degree):
        with open(filename,"w") as fw:
            # first the class header stuff
            fw.write("'''\n")
            fw.write("# Automaticaly generated InvariantVandermonde matrix\n")
            fw.write("# Generated from leuci-pol interpolation loibrary\n")
            fw.write("# (c) Rachel Alcraft, 2023, Birkbeck College, London University\n")
            fw.write("'''\n\n")
            fw.write("import numpy as np\n")
            fw.write(f"class InvariantVandermonde(object):\n")
            fw.write("\tdef __init__(self):\n")            
            fw.write("\t\tself.make_mat()\n")                         
            fw.write("\tdef get_invariant(self):\n")
            fw.write("\t\treturn self.alcraft\n")
            shx,shy = self.alcraft.shape
            fw.write("\tdef make_mat(self):\n")
            fw.write(f"\t\tself.alcraft = np.zeros(({shx},{shy}))\n")            
            for i in range(shx):
                for j in range(shy):
                    v = self.alcraft[i,j]
                    #self.invariant[0, 0] = 1.0
                    fw.write(f"\t\tself.alcraft[{i},{j}]={v}\n")



    


"""


"""

class Atom(object):
    def __init__(self, line):
        Line = line
        AtomNo = line.Substring(6, 5).Trim()
        AtomType = line.Substring(12, 4).Trim()
        occupant = line.Substring(16, 1).Trim()
        AA = line.Substring(17, 3).Trim()
        Chain = line.Substring(21, 1).Trim()
        ResidueNo = line.Substring(22, 5).Trim()
        insertion = line.Substring(26, 1).Trim()
        strX = line.Substring(30, 8).Trim()
        strY = line.Substring(38, 8).Trim()
        strZ = line.Substring(46, 8).Trim()
        X = Convert.ToDouble(strX)
        Y = Convert.ToDouble(strY)
        Z = Convert.ToDouble(strZ)
        Occupancy = Convert.ToDouble(line.Substring(54, 6).Trim())
        BFactor = Convert.ToDouble(line.Substring(60, 6).Trim())
        Element = line.Substring(66).Trim()
        Summary = ((((((Chain + ":") + ResidueNo) + ":") + AA) + ":") + AtomType)
    def distance(self, atm2):
        v1 = VectorThree(X, Y, Z)
        v2 = VectorThree(atm2.X, atm2.Y, atm2.Z)
        return v1.distance(v2)
    def coords(self):
        return VectorThree(X, Y, Z)



"""
RSA 25/2/23

This class handles 3d matrices
"""

import math
import numpy as np

# class interface
from maptial.xyz import vectorthree as v3

class MatrixThreeX(object):
    def __init__(self, vals = [0,0,0,0,0,0,0,0,0]):
        self.matrix = vals
                
    def get_inverse(self):                
        detWhole = self.get_determinant()
        transp = []
        transp.append(self.matrix[0])
        transp.append(self.matrix[3])
        transp.append(self.matrix[6])
        transp.append(self.matrix[1])
        transp.append(self.matrix[4])
        transp.append(self.matrix[7])
        transp.append(self.matrix[2])
        transp.append(self.matrix[5])
        transp.append(self.matrix[8])

        transpose = MatrixThreeX(transp)
        matinverse = MatrixThreeX()
        
        factor = 1;

        for i in range(3):        
            for j in range(3):                            
                detReduced = transpose.get_inner_determinant(i, j)
                #detReduced = self.get_inner_determinant(i, j)
                matinverse.put_value(detReduced * factor / detWhole, i, j)
                factor *= -1
        return matinverse
        
    def get_determinant(self):                
        factor = -1
        det = 0
        for i in range(3):        
            factor = factor * -1
            row_val = self.matrix[3 * i]
            newdet = self.get_inner_determinant(0, i)
            det = det + (factor * row_val * newdet)
        return det
    

    def get_inner_determinant(self, col, row):        
        smallMat = []
        if (col == 0):         
            if (row != 0):
                smallMat.append(self.matrix[1])
                smallMat.append(self.matrix[2])                
            if (row != 1):            
                smallMat.append(self.matrix[4])
                smallMat.append(self.matrix[5])            
            if (row != 2):            
                smallMat.append(self.matrix[7])
                smallMat.append(self.matrix[8])                        
        elif (col == 1):        
            if (row != 0):
                smallMat.append(self.matrix[0])
                smallMat.append(self.matrix[2])            
            if (row != 1):            
                smallMat.append(self.matrix[3])
                smallMat.append(self.matrix[5])            
            if (row != 2):            
                smallMat.append(self.matrix[6])
                smallMat.append(self.matrix[8])                    
        else: #(col == 1)        
            if (row != 0):
                smallMat.append(self.matrix[0])
                smallMat.append(self.matrix[1])            
            if (row != 1):            
                smallMat.append(self.matrix[3])
                smallMat.append(self.matrix[4])            
            if (row != 2):            
                smallMat.append(self.matrix[6])
                smallMat.append(self.matrix[7])                    
        n11 = smallMat[0]
        n12 = smallMat[1]
        n21 = smallMat[2]
        n22 = smallMat[3]
        return n11 * n22 - n12 * n21
        
    
    def get_value(self, row, col):                        
        pos = row * 3 + col
        return self.matrix[pos]
                            
    def put_value(self, val, row, col):    
        pos = row * 3 + col        
        self.matrix[pos] = val
        
    def multiply(self, col, byRow):        
        ######################################
        #So, this is by row not by column, or,,, anyway which is which...
        col0 = col.A
        col1 = col.B
        col2 = col.C
        scaled = v3.VectorThree()

        if byRow:
            poses = [0,1,2,3,4,5,6,7,8]
        else:
            poses = [0,3,6,1,4,7,2,5,8]
        s0 = col0 * self.matrix[poses[0]]
        s1 = col0 * self.matrix[poses[1]]
        s2 = col0 * self.matrix[poses[2]]
        s0 += col1 * self.matrix[poses[3]]
        s1 += col1 * self.matrix[poses[4]]
        s2 += col1 * self.matrix[poses[5]]
        s0 += col2 * self.matrix[poses[6]]
        s1 += col2 * self.matrix[poses[7]]
        s2 += col2 * self.matrix[poses[8]]
        scaled.put_by_idx(0, s0)
        scaled.put_by_idx(1, s1)
        scaled.put_by_idx(2, s2)
        return scaled

    def get_key(self,rnd=-1):
        if rnd > -1:
             return [round(x,rnd) for x in self.matrix]
        else:
            return str(self.matrix)

    
        



from os.path import exists
import urllib.request

from Bio.PDB.MMCIFParser import MMCIFParser        
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.PDBParser import PDBParser
import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

from maptial.geo import pdbobject as po

"""~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
class PdbLoader(object):
    def __init__(self, pdb_code, directory="", cif=False,source="ebi"):        
        self.pdb_code = pdb_code
        self.directory = directory        
        self.cif = cif
        self.pobj = po.PdbObject(pdb_code)                
        if source == "ebi":
            self.cif_filepath = f"{directory}{pdb_code}.cif"
            self.cif_url = f"https://www.ebi.ac.uk/pdbe/entry-files/download/{pdb_code}.cif"        
            self.pdb_filepath = f"{directory}{pdb_code}.pdb"
            self.pdb_url = f"https://www.ebi.ac.uk/pdbe/entry-files/download/pdb{pdb_code}.ent"
        elif source == "alphafold":
            #https://alphafold.ebi.ac.uk/files/AF-Q8CGX5-F1-model_v4.pdb
            #https://alphafold.ebi.ac.uk/files/AF-Q8CGX5-F1-model_v4.cif
            self.cif_filepath = f"{directory}{pdb_code}.cif"
            self.cif_url = f"https://alphafold.ebi.ac.uk/files/{pdb_code}.cif"
            self.pdb_filepath = f"{directory}{pdb_code}.pdb"
            self.pdb_url = f"https://alphafold.ebi.ac.uk/files/{pdb_code}.pdb"
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
    def exists_pdb(self):
        if exists(self._filepath):
            return True
        else:            
            return False
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
    def download_pdb(self,cif=False):
        if cif:
            if not exists(self.cif_filepath):
                try:            
                    urllib.request.urlretrieve(self.cif_url, self.cif_filepath)                                
                except:            
                    return False
        else:
            if not exists(self.pdb_filepath):
                try:            
                    urllib.request.urlretrieve(self.pdb_url, self.pdb_filepath)                                
                except:            
                    return False
        return True
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
    def load_pdb(self):
        loaded = False
        if self.cif:
            try:
                self.download_pdb(cif=True)
                structure = MMCIFParser().get_structure(self.pdb_code, self.cif_filepath)            
                loaded = True
            except Exception as e:                              
                print("Error loading cif file", str(e))
                loaded = False
        if not loaded:
            self.download_pdb(cif=False)            
            structure = PDBParser(PERMISSIVE=True).get_structure(self.pdb_code, self.pdb_filepath)                                        
        self.pobj.add_atoms(structure)          
        #print(structure.header)
        return self.pobj
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""


"""
RSA 25/2/23

This class handles 3d vectors
"""

import math
import numpy as np

# class interface

class VectorThree(object):
    def __init__(self, a=0.,b=0.,c=0.,abc=[0.,0.,0.]):    
        if a != 0 or b != 0 or c!=0:
            self.A = a
            self.B = b
            self.C = c        
            self.Valid = True
            self.npy = np.zeros((3))
            self.npy[0] = a
            self.npy[1] = b
            self.npy[2] = c
        else:
            self.npy = np.array(abc)        
            self.A = self.npy[0]
            self.B = self.npy[1]
            self.C = self.npy[2]



    
    def from_coords(self, coords):    
        coords = coords.strip()
        if coords[0] == "(":
            coords = coords[1:]
        if coords[-1] == ")":
            coords = coords[:-1]
        a,b,c = coords.split(",")
        self.A = float(a)
        self.B = float(b)
        self.C = float(c) 
        self.Valid = True
        self.npy = np.array([self.A,self.B,self.C])
        return VectorThree(abc=self.npy.tolist())
                        
    def make_from_key(self,key):                            
        key = key.Substring(1)
        key = key.Substring(0,key.Length-1)
        sk = key.Split(",")
        self.A = float(sk[0])
        self.B = float(sk[1])
        self.C = float(sk[2])        
        self.npy = np.array([self.A,self.B,self.C])
        self.Valid = True
        
    def get_by_idx(self,idx):    
        return  self.npy[idx]   
        #if idx == 0:
        #    return self.A;
        #elif idx == 1:
        #    return self.B;
        #else:
        #    return self.C;
        
    def put_by_idx(self, idx, val):
        self.npy[idx] = val
        if idx == 0:
            self.A = val            
        elif idx == 1:
            self.B = val
        else:
            self.C = val
        
    def distance(self,ABC):    
        dis = (self.A - ABC.A) * (self.A - ABC.A) + (self.B - ABC.B) * (self.B - ABC.B) + (self.C - ABC.C) * (self.C - ABC.C)
        return math.sqrt(dis)
    
    def magnitude(self):        
        mag = (self.A * self.A) + (self.B * self.B) + (self.C * self.C)
        return math.sqrt(mag)
                                                
    def get_angle(self, ABC):    
        BA = VectorThree(0 - self.A, 0 - self.B, 0 - self.C)
        BC = VectorThree(0 - ABC.A, 0 - ABC.B, 0 - ABC.C);
        dot = BA.dot_product(BC);
        magBA = BA.magnitude();
        magBC = BC.magnitude();
        cosTheta = dot / (magBA * magBC);
        theta = math.acos(cosTheta);
        return theta #in radians
    
    def dot_product(self, ABC):
        px = self.A * ABC.A
        py = self.B * ABC.B
        pz = self.C * ABC.C
        return px + py + pz
    
    def get_key(self,rnd = 4):
        strkey = "(" + str(round(self.A, rnd)) + ","
        strkey += str(round(self.B, rnd)) + ","  
        strkey += str(round(self.C, rnd)) + ")"  
        return strkey
            
    def get_point_pos(self, samples, width):    
        gap = width/(samples-1)                        
        PP = VectorThree(self.A, self.B, self.C)        
        PP.A = PP.A / gap
        PP.B = PP.B / gap
        PP.C = PP.C / gap
        adj = (samples-1)/2#(width / (2 * gap))
        #if (int)(samples % 2) != 0:
        #    adj -= 0.5
        PP.A += adj
        PP.B += adj                
        return PP

# static functions

def v3_add(ABC, PQR):        
    A = ABC.A + PQR.A
    B = ABC.B + PQR.B
    C = ABC.C + PQR.C
    return VectorThree(A,B,C)
    
def v3_subtract(ABC, PQR):        
    A = ABC.A - PQR.A
    B = ABC.B - PQR.B
    C = ABC.C - PQR.C
    return VectorThree(A,B,C)

def v3_divide(ABC, PQR):        
    A = ABC.A / PQR.A
    B = ABC.B / PQR.B
    C = ABC.C / PQR.C
    return VectorThree(A,B,C)

def v3_multiply(ABC, PQR):        
    A = ABC.A * PQR.A
    B = ABC.B * PQR.B
    C = ABC.C * PQR.C
    return VectorThree(A,B,C)
